%% graph.pl :: 
%%	genera per tot subgraf independent, llista constraints amb llista cicles estrictes
%%	Si opci de cancel.lar cicles : disjunci de possibles solucions


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% funcions de lligam mspo - constraints
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% retorna:
%   [ [NumConstraintMSPO-1, LcyclesOut1],[NumConstraintMSPO-2, LcyclesOut2],...]
%   NumConstraintMSPO cal per mostrar soluci	 
%   LcyclesOut= [LcycleSubgraph1, LcycleSubgraph2,...]
%
% NORMAL no deeper_analysis
%   LcycleSubgraf=[QenumENcicles,llista_cicles_amb_estricte,llista_cicles_sense_estricte]
%       e.g. QenumENcicles=[q(1,->-),q(2,->=-),...], llista_cicles_amb_estricte=[[1,2],[2,3,4],...]
%	     llista_cicles_sense_estricte --- noms cal per mostrar soluci
%
% INNERMOST  o deeper_analysis  (TRY_CANCEL_CYCLES):
%   LcycleSubgraf=[Qenum,LStrictCond,LNoStrictCond,[OR1Subgraf,OR2Subgraf,...]] 	  
%		   Qenum=[q(1,->-),q(2,->=-),...]
%		   LStrictCond-LNoStrictCond==[[Lc1,1,2],[Lc2,2,3,4],...]
% primera manera:
%   OR1Subgraf=[K,QenumENcicles,llista_cicles_cond_estricte]
%	e.g. QenumENcycles=[1,2,3,4] 
%	     llista_cicles_cond_estricte=[[Lc1,1,2],[Lc2,2,3,4],...]
%	-K s 'a1' quan tot estrictes i en option_1 (QenumENcicles s [] en option_1)
%	-K s 'a2' en option_2 quan hi ha estrictes i no-estrictes i ens quedem amb 
%	 els estrictes sense condicions: llista_cicles_cond_estricte=[[[],1,2],[[],2,3,4],...]
%
% segona manera: 'b' - noms si DEEPER_ANALYSIS	(poden haver-hi varis 'b' (s un OR)) - option_3
%   OR1Subgraf=['b','i',l-condici-1,l-cancelled-cyc,ConjuntSubgraphIntern-1]   
%     i : identifica el 'b' dins el OR1Subgraf (per mostrar soluci)
%     l-condici-1 = llista condicions que cancel.la el cicle
%     l-cancelled-cyc = [L-cancelled-cyc-Strict,L-cancelled-cyc-NOStrict]
%	format: [numscyc1,numscyc2,...], e.g. [[1,2],[2,3,4],...]
% 	n'hi pot haver varis perqu per diferents condicions el subgraf pot ser diferent (conds compartides).
%     ConjuntSubgraphIntern= [LcycleSubgraphIntern1, LcycleSubgraphIntern2,...]
%	LcycleSubgraphIntern=[K,NumsEncicles,l_cycs_cond_strict,l_cycs_NOstrict] (cas 'a1' o 'a2')
%	- cas a1 i a2 com abans per afegint un quart param que servir per mostrar la soluci
%	- [a1,QenumEncicles,llista_cycles_cond_estricte,[]]
%	- [a2,QenumEncicles,llista_cycles_NOcond_estricte,llista_cycles_NOcond_NoStrict]

list_const_cycles(Rules,LC,List_C_Cycles):-
	list_const_cycles(Rules,LC,[],List_C_Cycles),!.

list_const_cycles(_,[],L,L):-!.
list_const_cycles(Rules,[[N,Q]|R],L,LOut):-
	convert(Q,Q1),
	cycles(Rules,Q1,Qenum,Lcycles),!,   
% Lcycles=[Subg1,Subg2,...], 
% Cas NORMAL SubgI=[AllNums,Lnumscycstrict,Lnumscycsnostrict]
% Cas TRY_CANCEL_CYCLES SubgI=[AllNums,Lnumscycstrict,Lnumscycsnostrict,[S1,S2,...]] Sj=[K,..] K='a1' or 'a2' or 'b'
	set_Qenum_each_Subgraph(Qenum,Lcycles,LcyclesOut),
	list_const_cycles(Rules,R,[[N,LcyclesOut]|L],LOut).

convert([],[]):-!.
convert([X|R],[q(X,[])|R1]):- convert(R,R1),!.

% converteix les constants x,y,z,x1,x2,... que hi hagi a la llista
% de constraints generades per mspo en variables.

conv_cts_vars(LC,LC1):-
		variables(Lconst),
		llvarsnews(Lconst,Lvars),
		replace(Lvars,LC,LC1).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Substitueix la llista de nums de constraints de cada subgraf pels Qenum corresponents
% Subgraph=[List-NumsSubgraph,List-StrictCycles,List-NoStrictCycles] (normal)
% Subgraph=[List-NumsSubgraph,List-StrictCycles,List-NoStrictCycles,ORsols] (inner o deeper)
set_Qenum_each_Subgraph(_,[],[]).
set_Qenum_each_Subgraph(Qenum,[[NumsSubgraph|R]|Lcycles],[[QSubgraph|R]|LcyclesOut]):-
	selec_nums_Qenum_and_renQenum(NumsSubgraph,Qenum,QSubgraph),!,
	set_Qenum_each_Subgraph(Qenum,Lcycles,LcyclesOut),!.


selec_nums_Qenum_and_renQenum([],_,[]).
selec_nums_Qenum_and_renQenum([Num|NumsSubgraph],Qenum,[q(Num,C1)|QenumNums]):-
	member21(q(Num,C,_),Qenum),
	renvars(C,C1),
	selec_nums_Qenum_and_renQenum(NumsSubgraph,Qenum,QenumNums),!.
	
%simplify_and_renQenum([],[]):-!.
%simplify_and_renQenum([q(N,C,_)|Qenum],[q(N,C1)|Qenum1]):-
%	renvars(C,C1),
%	simplify_and_renQenum(Qenum,Qenum1),!.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

cycles(Rules,Q,Q1,Lcycles):-
	leftR0(Def),     
	rentops_i_enum(Q,1,Q1),
	abstraction(Q1,Def,Qabs), 
	build_and_simplify_graph(Rules,Qabs,SubGraphs),
  %SubGraphs=[ [[q(N,CONS,Lc),List_connexions],...] ...]
	gen_cycles(Rules,SubGraphs,Lcycles),!.


%% ------------------------------------------------------------
%% renombra els tops pel mateix smbol amb majscula (estil DP)
%% ------------------------------------------------------------

rentops_i_enum([],_,[]).
rentops_i_enum([X|R],N,[X1|R1]):-rentop(X,N,X1), N1 is N+1, rentops_i_enum(R,N1,R1).

rentop(q(C,L),N,q(N,C1,L)):-C=..[Op|[S,T]],
			rentop1(S,S1),
			rentop1(T,T1),
			C1=..[Op|[S1,T1]],!.

rentop1(T,T1):-	T=..[F|Args],
		name(F,X),X=[X1|R],
		( (X1>=97,X1=<122,!, Y is X1-32,New=[Y|R])
		  ;
		  (New=[39,X1,39|R])
		),!, 
		name(F1,New),
		T1=..[F1|Args],!.

%% -------------------------------------------------------------------------------------
%% fa l'abstracci a la part dreta dels parells posant una nova variable 
%% enlloc dels subtermes amb top definit i refresca les variables 
%% (si no estem en cas innermost).
%% -------------------------------------------------------------------------------------

abstraction(Q,Def,Qabs):-
	abstraction(Q,Def,[],Qabs).

abstraction([],_,L,L):-!.
abstraction([q(N,C,Lc)|R],Def,Labs,Qabs):-
		abst(q(N,C,Lc),Def,Cabs),
		concat312(Labs,[Cabs],L),
		abstraction(R,Def,L,Qabs),!.

abst(q(N,C,Lc),Def,q(N,C1,Lc11)):-
		C=..[Op|[S,T]],
		cap_ren0(S,T,Def,T1,Lc,Lc1),
		nf_if_inn(S,Lc1,Lc11),
		C1=..[Op|[S,T1]].


cap_ren0(S,T,Def,TOut,Lc,LcOut):-
	innermost,!,
	cap_ren_inn(S,T,Def,TOut,Lc,LcOut),!.
cap_ren0(_,T,Def,TOut,Lc,LcOut):-
	cap_ren(T,Def,TOut,Lc,LcOut),!.

cap_ren(T,_,T1,L,[T>=T1|L]):-var(T),!.
cap_ren(T,Def,T1,L,[T>=T1|L]):-T=..[F|_],
			     member21([F,_],Def),!.
cap_ren(T,Def,T1,Lc,L):-T=..[F|A],
			l_cap_ren(A,Def,Lc,L,A1),
			T1=..[F|A1].

l_cap_ren([],_,L,L,[]):-!.
l_cap_ren([X|R],Def,Lc,L,[X1|R1]):-cap_ren(X,Def,X1,Lc,L1),
				   l_cap_ren(R,Def,L1,L,R1),!.


% no renombra variables ni canvia per variable nova els subtermes 
% de t amb top definit que tamb ho sn de s.
cap_ren_inn(_,T,_,T,L,L):-var(T),!.
cap_ren_inn(S,T,Def,TOut,L,LOut):-T=..[F|_],
			     member21([F,_],Def),
			     (( member_term2(T,S),!,
				TOut=T, LOut=L )
	 		      ;
			      ( LOut=[T>=TOut|L] )
			     ),!.
cap_ren_inn(S,T,Def,T1,Lc,L):-T=..[F|A],
			l_cap_ren_inn(S,A,Def,Lc,L,A1),
			T1=..[F|A1].

l_cap_ren_inn(_,[],_,L,L,[]):-!.
l_cap_ren_inn(S,[X|R],Def,Lc,L,[X1|R1]):-cap_ren_inn(S,X,Def,X1,Lc,L1),
				       l_cap_ren_inn(S,R,Def,L1,L,R1),!.


nf_if_inn(S,Lc,LcOut):-
	innermost,!,
	S=..[_|Args],
	add_nf(Args,Lc,LcOut),!.
nf_if_inn(_,Lc,Lc):-!.

add_nf([],L,L):-!.
add_nf([X|R],L,LOut):-
	addifno1(nf(X),L,L1),
	add_nf(R,L1,LOut),!.


%s'usa des de nf.pl
add_inn(L,Lc,LcOut):-
	innermost,!,
	L=..[_|Args],
	add_nf(Args,Lc,LcOut),!.
add_inn(_,Lc,Lc).

%% -------------------------------------------------------------------------------------
%%  Construcci del graph aproximat com fan els D.P. 
%%  (considerant cada node independentment) ( normal o innermost )
%%  Simplificar nodes que no pertanyen a cap cicle.
%% -------------------------------------------------------------------------------------

build_and_simplify_graph(Rules,Qabs,SubGraphs):-
	build_graph(Rules,Qabs,Qabs,Graph),
	simplify_graph(Graph,SubGraphs),!.

build_graph(_,[],_,[]).
build_graph(Rules,[q(N,Q,Lc)|Qabs],AllQ,[[q(N,Q,Lc),LM]|Graph]):-
	findall(M,connect_graph(Rules,Q,AllQ,M),LM),
	build_graph(Rules,Qabs,AllQ,Graph),!.

connect_graph(Rules,Q1,[q(N,Q2,_)|_],N):-
	Q1=..[_|[_,R]],
	Q2=..[_|[L,_]],
	refreshterm(R,R1),
	refreshterm(L,L1),
	unify0(R1,L1), check_sat_nf_args(Rules, L1).	
connect_graph(Rules,Q1,[_|Rest],N):-
	connect_graph(Rules,Q1,Rest,N).


unify0(S,T):-innermost,!,unifica(S,T),!.
unify0(S,T):-S=T,!.

% comprova que les condicions de NF dels arguments del terme esquerra
% del constraint on hem connectat es satisfacin.
check_sat_nf_args(_,_):-
	\+ innermost,!.  % si  no s innermost no hi ha nf()
check_sat_nf_args(Rules,T):-
	T=..[_|Args],
	remove_variables_and_renvars(Args,[],ArgsNF),
	sat_nf(Rules,ArgsNF),!.

remove_variables_and_renvars([],L,L).
remove_variables_and_renvars([X|Args],L,ArgsNF):-
	var(X),!,	
	remove_variables_and_renvars(Args,L,ArgsNF).
remove_variables_and_renvars([X|Args],L,ArgsNF):-
	renvars(X,X1),
	addifno(nf(X1),L,L1),
	remove_variables_and_renvars(Args,L1,ArgsNF).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Simplification of the graph
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

simplify_graph(Graph,Subgraphs):-
	simplify_graph0(Graph,Graph1),!,
	remove_subgraphs_allge(Graph1,Subgraphs),!.

simplify_graph0(Graph,GraphOut):-
	simplify_graph1(Graph,Graph1),!,
	remove_no_return(Graph1,Graph2),
	( (Graph1==Graph2,!,GraphOut=Graph2)
	  ;
	  (simplify_graph0(Graph2,GraphOut))
	),!.


simplify_graph1(Graph,GraphOut):-
	remove_nowhere(Graph,Graph1),
	remove_itself_only(Graph1,Graph2,[]),
	((Graph==Graph2,!,GraphOut=Graph2)
	 ;
	 (simplify_graph1(Graph2,GraphOut))
	),!.

remove_nowhere(Graph,GraphOut):-
	delete33([q(N,_,_),[]],Graph,Graph1),!,
	delete_N_graph(N,Graph1,Graph2),
	remove_nowhere(Graph2,GraphOut),!.
remove_nowhere(Graph,Graph).

	
% removes M in all lists LM for each node [N,LM]
delete_N_graph(_,[],[]):-!.
delete_N_graph(M,[[q(N,Q,L),LM]|Rest],[[q(N,Q,L),LM1]|Rest1]):-
	delete0312(M,LM,LM1),!,
	delete_N_graph(M,Rest,Rest1),!.

remove_itself_only(Graph,GraphOut,Ldone):-
	member22([q(N,_,_),[N]],Graph),	% ho faig aix per no canviar ordre del graf ja que
	\+(member21(N,Ldone)),!,	% per saber si ha acabat la simplificaci els comparo.
	delete_N_others_graph(N,Graph,Graph1),  % (veure simplify_graph)
	remove_itself_only(Graph1,GraphOut,[N|Ldone]),!.
remove_itself_only(Graph,Graph,_).

% esborra M de la llista de connexions de tots els nodes del graph excepte ell mateix
% Pel node M, si s >= l'elimina i sin el deixa igual.
delete_N_others_graph(_,[],[]):-!.
delete_N_others_graph(M,[[q(M,_>=_,_),_]|Rest],Rest1):-!,
	delete_N_others_graph(M,Rest,Rest1),!.
delete_N_others_graph(M,[[q(M,S>T,Lc),LM]|Rest],[[q(M,S>T,Lc),LM]|Rest1]):-!,
	delete_N_others_graph(M,Rest,Rest1),!.
delete_N_others_graph(M,[[q(N,Q,Lc),LM]|Rest],[[q(N,Q,Lc),LM1]|Rest1]):-!,
	delete0312(M,LM,LM1),!,
	delete_N_others_graph(M,Rest,Rest1),!.


remove_no_return(Graph,GraphOut):-
	findall([N,LN],member22([q(N,_,_),LN],Graph),Allcon),
	close_connexions(Allcon,AllconTrans),
	simplify_no_return(AllconTrans,Graph,GraphOut),!.

close_connexions(AllConn,AllconTrans):-
	delete33([N,LN],AllConn,AllConn1),
	delete0312(N,LN,LN1),
	obtain_conn_from(LN1,AllConn1,[],LM1),
	difer(LM1,LN,Dif),Dif=[_|_],!,
	union(LN,LM1,LN2),
	concat312(AllConn1,[[N,LN2]],AllConn2),
	close_connexions(AllConn2,AllconTrans),!.
close_connexions(AllConn,AllConn).
	
obtain_conn_from([],_,LM0,LM0).
obtain_conn_from([N|LN],AllConn,LM0,LMOut):-
	member21([N,L],AllConn),
	union(L,LM0,LM1),
	obtain_conn_from(LN,AllConn,LM1,LMOut),!.


simplify_no_return(_,[],[]).
simplify_no_return(ConnTrans,[[q(N,Q,Lc),LN]|Graph],[[q(N,Q,Lc),LN1]|GraphOut]):-
	conn_can_return(N,LN,ConnTrans,LN1),
	simplify_no_return(ConnTrans,Graph,GraphOut),!.
	

conn_can_return(_,[],_,[]):-!.
conn_can_return(N,[N|LN],ConnTrans,[N|LNOut]):-!,
	conn_can_return(N,LN,ConnTrans,LNOut),!.
conn_can_return(N,[M|LN],ConnTrans,[M|LNOut]):-
	member21([M,LM],ConnTrans),
	member21(N,LM),!,
	conn_can_return(N,LN,ConnTrans,LNOut),!.
conn_can_return(N,[_|LN],ConnTrans,LNOut):-!,
	conn_can_return(N,LN,ConnTrans,LNOut),!.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

remove_subgraphs_allge(Graph,SubGraphsOut):-
	findall([N,LN],member22([q(N,_,_),LN],Graph),Allcon),
	close_connexions(Allcon,AllconTrans),
	obtain_subgraphs(AllconTrans,NumSubgraphs),
	remove_Sub_allge(NumSubgraphs,Graph,SubGraphsOut),!.

obtain_subgraphs([],[]).
obtain_subgraphs([X|AllconTrans],[Subgraph|SubgraphsOut]):-
	obtain_subgraph([X|AllconTrans],AllconTrans1,Subgraph),!,
	obtain_subgraphs(AllconTrans1,SubgraphsOut),!.

obtain_subgraph([[N,LN]|AllconTrans],AllconTransOut,Subgraph):-
	delete032(N,LN,LN1),
	obtain_subgraph_N([N],LN1,AllconTrans,AllconTransOut,Subgraph),!.
	
obtain_subgraph_N(Lsubgraph,[],AllconTrans,AllconTrans,Lsubgraph):-!.
obtain_subgraph_N(Lsubgraph,LNsub,AllconTrans,AllconTransOut,Subgraph):-
	get_LNsub(LNsub,AllconTrans,AllconTrans1,[],LconSub),
	concat312(Lsubgraph,LNsub,Lsubgraph1),
	difer(LconSub,Lsubgraph1,LOthers),
	obtain_subgraph_N(Lsubgraph1,LOthers,AllconTrans1,AllconTransOut,Subgraph),!.

get_LNsub([],AllconTrans,AllconTrans,L,L).
get_LNsub([N|LNsub],AllconTrans,AllconTransOut,L0,LOut):-
	delete32([N,LN],AllconTrans,AllconTrans1),!,
	union(L0,LN,L1),
	get_LNsub(LNsub,AllconTrans1,AllconTransOut,L1,LOut),!.
get_LNsub([_|LNsub],AllconTrans,AllconTransOut,L0,LOut):-
	get_LNsub(LNsub,AllconTrans,AllconTransOut,L0,LOut),!.

	
remove_Sub_allge([],_,[]).
remove_Sub_allge([LNum|NumSubgraphs],Graph,[SubG1|SubGraphOut]):-
	member22(N,LNum),
	member21([q(N,_>_,_),_],Graph),!,
	remove_LNum_Graph(LNum,Graph,SubG1,Graph1),!,
	remove_Sub_allge(NumSubgraphs,Graph1,SubGraphOut),!.
remove_Sub_allge([LNum|NumSubgraphs],Graph,SubGraphOut):-
	remove_LNum_Graph(LNum,Graph,_,Graph1),!,
	remove_Sub_allge(NumSubgraphs,Graph1,SubGraphOut),!.


remove_LNum_Graph([],Graph,[],Graph).
remove_LNum_Graph([N|LNum],Graph,[[q(N,C,Lc),Con]|Rest],GraphOut):-
	delete32([q(N,C,Lc),Con],Graph,Graph1),!,
	remove_LNum_Graph(LNum,Graph1,Rest,GraphOut),!.
	
	

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

obt_lc([],Lc,Lc):-!.
obt_lc([q(_,_,Lcq)|Rest],Lc,LcOut):-
	sel_poss_nocycle(Lcq,Lc1),
	concat312(Lc,Lc1,Lc11),
	obt_lc(Rest,Lc11,LcOut),!.

set_new_vars([]):-!.
set_new_vars(NewRules):-
	retract(variables(Vars)), 
	obtain_vars_system(NewRules,VarsNewRules),
	length21(Vars,L1), length21(VarsNewRules,L2),
	( ( L1>=L2,!,assert(variables(Vars))) 
	  ;
	  (modificar_variables(Vars,VarsNewRules))
	),!.
	
modificar_variables(Vars,VarsNew):-
	member21(X,Vars),  % tots o cap comencen amb _
	name(X,[95|_]),!,
	change_name_vars(VarsNew,Vars1),
	assert(variables(Vars1)),!.
modificar_variables(_,VarsNew):-
	assert(variables(VarsNew)),!.

sel_poss_nocycle([],[]):-!.
sel_poss_nocycle([C|Rest],[C|Rest1]):-
	to_consider(C),!,
	sel_poss_nocycle(Rest,Rest1),!.
sel_poss_nocycle([_|Rest],Rest1):-
	sel_poss_nocycle(Rest,Rest1),!.

to_consider(nf(X)):-var(X),!,fail.
to_consider(nf(_)):-!.
to_consider(td(_,_)):-!.
to_consider(S>=T):-
	var(S),!, member_term2(S,T),!.
to_consider(S>=T):-
	nonvar(T),\+(S==T),!.
%to_consider(_>T):-
%	nonvar(T),!.
%to_consider(S>T):-
%	var(T), member_term2(T,S),!.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% ------------------------------------------------------------------------
%% genera tots els cicles simples (cam i condicions) a partir del graph.
%% ------------------------------------------------------------------------

gen_cycles(_,[],[]).
gen_cycles(Rules,[Subgraph|Rest],LOut):-
	gen_cycles1(Rules,Subgraph,LL1),!,
	gen_cycles(Rules,Rest,LRest),
	concat312(LL1,LRest,LOut),!.		

% tractament d'un subgraf1 (encara pot subdividir-se en subgrafs)
% LSUBSOLCycles=[Subgraf11,Subgraf12,...]
gen_cycles1(Rules,Graph,LSUBSOLCycles):-
	try_cancel_cycles,!,     %innermost o deeper analysis
	obtain_simple_cycles_from_graph(Graph,NumCycles),
	obtain_subgraphs_and_conditions_path_cycles(Rules,Graph,NumCycles,L_SCycles),
% els cicles que no es compleix NF o no pas reescr. ja s'han eliminat i no surten als
% elements de L_SCycles. L_SCycles=[SCyclesSUB1,SCyclesSUB2,...] (un elem per subgraf)
% SCyclesSUBi= list of [list-conditions,nums-cycle]. e.g. [ [[cond1,cond2,...],1,3,4], ...] 
% contain all cycles with o without stricts
	generate_constraints_cycles_each_subgraph(Rules,Graph,L_SCycles,LSUBSOLCycles),!. 

% amb reescriptura normal sense opci cancel-cycles, tots els constraints que
% surten a Graph sn exactament els mateixos que surten a NumCycles
gen_cycles1(_,Graph,[[NumsGraph,StrictCycles,NonStrictCycles]]):-
	obtain_simple_cycles_from_graph(Graph,NumCycles),
	separ_strict_nonstrict_cycles(NumCycles,Graph,StrictCycles,NonStrictCycles),   
	map(numsGraph,Graph,NumsGraph),!.	

numsGraph([q(N,_,_),_],N).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

obtain_simple_cycles_from_graph(Graph,NumCyclesOut):-
	findall([N,LN],member22([q(N,_,_),LN],Graph),Conn_Graph),	
	map(head,Conn_Graph,Lnums),
	obtain_simple_cycles(Lnums,Conn_Graph,NumCyclesOut),!.

	
% Graph cont noms les connexions (no els constraints...)
% S'obtenen a NumCycles tots els cicles simples: sense repetir cap node
obtain_simple_cycles([],_,[]).
obtain_simple_cycles([N|Lnums],Graph,NumsCycles):-
		% construeix tots els cicles que comencen per (contenen) N
	cons_cycles(N,[[N]],Graph,[],C1),!, 
	remove_N_Graph(N,Graph,Graph1),	
	obtain_simple_cycles(Lnums,Graph1,C2),
	concat312(C1,C2,NumsCycles),!.
	
remove_N_Graph(_,[],[]).
remove_N_Graph(N,[[N,_]|Rest],Rest1):-!,
	remove_N_Graph(N,Rest,Rest1),!.
remove_N_Graph(N,[[M,LM]|Rest],[[M,LM1]|Rest1]):-
	delete0312(N,LM,LM1),!,
	remove_N_Graph(N,Rest,Rest1),!.


cons_cycles(_,[],_,L,L):-!.
cons_cycles(N,[[H|Cam]|Rcams],G,LCycles,LCyclesOut):-
	member21([H,LH],G),
	( ( member21(N,LH),!,reverse1([H|Cam],Cyc1),C1=[Cyc1|LCycles] )
	  ;
	  ( C1=LCycles )
	),!,
	difer(LH,[H|Cam],LH1),
	map(amplia_cami,[H|Cam],LH1,NewCams),
	concat312(Rcams,NewCams,Lcams),!,
	cons_cycles(N,Lcams,G,C1,LCyclesOut),!.

amplia_cami(Cam,X,[X|Cam]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

separ_strict_nonstrict_cycles(NumCycles,Graph,StrictCycles,NonStrictCycles):-   
	findall(N,member22([q(N,_>_,_),_],Graph),LStricts),!,
	separ_strict_nonstrict_cycles1(NumCycles,LStricts,StrictCycles,NonStrictCycles),!.
	
separ_strict_nonstrict_cycles1([],_,[],[]).
separ_strict_nonstrict_cycles1([Nums|NumCycles],LStricts,[Nums|StrictCycles],NonStrictCycles):-
	inter(Nums,LStricts,[_|_]),!,	 
	separ_strict_nonstrict_cycles1(NumCycles,LStricts,StrictCycles,NonStrictCycles),!.
separ_strict_nonstrict_cycles1([Nums|NumCycles],LStricts,StrictCycles,[Nums|NonStrictCycles]):-
	separ_strict_nonstrict_cycles1(NumCycles,LStricts,StrictCycles,NonStrictCycles),!.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% estem usant opci de cancel.lar cicles (deeper-analysis o innermost)

obtain_subgraphs_and_conditions_path_cycles(_,_,[],[]):-!.
obtain_subgraphs_and_conditions_path_cycles(Rules,Graph,NumsCycles,LSUBConds):-
	remove_no_real_cycles(Rules,Graph,NumsCycles,NumsCycles1),
	findall(N,member22([q(N,_>_,_),_],Graph),LStricts),
	obtain_new_subgraphs(LStricts,NumsCycles1,LSUBCycles),
	obtain_conditions_path_cycles_each_subgraph(Rules,Graph,LSUBCycles,LSUBConds),!.

%%%%%%%%%%%%%
remove_no_real_cycles(Rules,Graph,NumsCycles,NumsCyclesOut):-
	selec_nodes_several_cycles(NumsCycles,Nodes),
	delete33(NumsCyc1,NumsCycles,NumsCycles1),
	check_no_real_cycle(Rules,Graph,Nodes,NumsCyc1),!,
	remove_no_real_cycles(Rules,Graph,NumsCycles1,NumsCyclesOut),!.
remove_no_real_cycles(_,_,NumsCycles,NumsCycles).
	
	
%%%
selec_nodes_several_cycles([],[]):-!.
selec_nodes_several_cycles(NumCycles,Nodes):-
	l_union(NumCycles,Allnodes),
	selec_nodes_several_cycles(Allnodes,NumCycles,Nodes),!.

selec_nodes_several_cycles([],_,[]):-!.
selec_nodes_several_cycles([N|AllNodes],NumCycles,[N|Nodes]):-
	get_cycle_contain_N(N,NumCycles,NumCycles1),
	get_cycle_contain_N(N,NumCycles1,_),!,
	selec_nodes_several_cycles(AllNodes,NumCycles,Nodes),!.
selec_nodes_several_cycles([_|AllNodes],NumCycles,Nodes):-
	selec_nodes_several_cycles(AllNodes,NumCycles,Nodes),!.

get_cycle_contain_N(N,Lnums,LnumsOut):-
	delete33(Nums,Lnums,LnumsOut),
	member21(N,Nums),!.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

check_no_real_cycle(Rules,Graph,Nodes,NumsCycle):-
	perform_cycle(Graph,Nodes,NumsCycle,Conditions),
	check_true_conds(Rules,Conditions),!,fail.
check_no_real_cycle(_,_,_,_):-!.   % no pot tancar cicle o no NF o no pas reescr.

check_true_conds(_,C):-
	\+ innermost,!,
	check_possible_reachability(C),!.
check_true_conds(Rules,C):-
	sel_and_sat_nf(Rules,C,C1),
	check_possible_reachability(C1),
	try_rewrite_step(Rules,C1,C2),
	check_possible_reachability(C2),!.

check_possible_reachability([]):-!.
check_possible_reachability(L):-
	sele_ge_left_novar(L,L1,_),
	del_repes(L1,L2),
	check_reachability(L2),!.

check_reachability([]):-!.
check_reachability([S>=T|Rest]):-
	leftR0(Def),
	check_reachability1(S,T,Def),!,
	check_reachability(Rest),!.
	
check_reachability1(S,_,_):- is_variable(S),!.
check_reachability1(_,T,_):- is_variable(T),!.
check_reachability1(S,_,Def):- 
	S=..[F|_],
	member21([F,_],Def),!.
check_reachability1(S,T,Def):- 
	S=..[F|ArgsS],
	T=..[G|ArgsT],
	F==G,
	check_reachability1_list(ArgsS,ArgsT,Def),!.
	
check_reachability1_list([],_,_):-!.
check_reachability1_list([X|Rx],[Y|Ry],Def):-
	check_reachability1(X,Y,Def),!,
	check_reachability1_list(Rx,Ry,Def),!.
 

obtain_conds_and_try_rewrite_step(Rules,Conditions,CondOut):-
	sele_ge_left_novar(Conditions,Cond,_),
	del_repes(Cond,Cond1),
	( ( innermost,!,
	    try_rewrite_step(Rules,Cond1,CondOut)
	  )
	  ;
	  (CondOut=Cond1)
	),!.
	  %mai falla perqu els que fallen ja els hem eliminat abans.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

obtain_new_subgraphs(_,[],[]):-!.
obtain_new_subgraphs(LStricts,[Nums|NumsCycles],ListSubgraphsCycles):-
	obtain_new_subgraphs1(NumsCycles,[Nums],ListGroups),
	separ_cycles_wrt_groups([Nums|NumsCycles],ListGroups,LSubgraphsCycles),
	elim_Sub_allge(LSubgraphsCycles,LStricts,ListSubgraphsCycles),!.

obtain_new_subgraphs1([],L,L):-!.
obtain_new_subgraphs1([Nums|NumsCycles],L,LOut):-
	obtain_group_nums(Nums,L,L1,LNums),
	obtain_new_subgraphs1(NumsCycles,[LNums|L1],LOut),!.

obtain_group_nums(Nums,L,LOut,LNums):-
	delete33(LN,L,L1),
	inter(Nums,LN,[_|_]),!,
	union(Nums,LN,Nums1),
	obtain_group_nums(Nums1,L1,LOut,LNums),!.
obtain_group_nums(Nums,L,L,Nums).

separ_cycles_wrt_groups([],_,[]):-!.
separ_cycles_wrt_groups(NumsCycles,[G|LG],[SubCycles|ListSubgraphsCycles]):-
	separ_cycles_1_group(NumsCycles,G,SubCycles,NumsCycles1),!,
	separ_cycles_wrt_groups(NumsCycles1,LG,ListSubgraphsCycles),!.

separ_cycles_1_group([],_,[],[]):-!.
separ_cycles_1_group([Nums|NumsCycles],G,[Nums|SubCycles],NumsCyclesOut):-
	inter(Nums,G,[_|_]),!,
	separ_cycles_1_group(NumsCycles,G,SubCycles,NumsCyclesOut),!.
separ_cycles_1_group([Nums|NumsCycles],G,SubCycles,[Nums|NumsCyclesOut]):-
	separ_cycles_1_group(NumsCycles,G,SubCycles,NumsCyclesOut),!.


elim_Sub_allge([],_,[]):-!.
elim_Sub_allge([Sub|LSubgraphsCycles],LStricts,[Sub|LSubgraphsCyclesOut]):-
	l_union(Sub,NumsSub),
	inter(NumsSub,LStricts,[_|_]),!,
	elim_Sub_allge(LSubgraphsCycles,LStricts,LSubgraphsCyclesOut),!.
elim_Sub_allge([_|LSubgraphsCycles],LStricts,LSubgraphsCyclesOut):-
	elim_Sub_allge(LSubgraphsCycles,LStricts,LSubgraphsCyclesOut),!.
	

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

obtain_conditions_path_cycles_each_subgraph(_,_,[],[]):-!.
obtain_conditions_path_cycles_each_subgraph(Rules,Graph,[SUBCycles|LSUBCycles],[SUBConds|LSUBConds]):-
	obtain_conditions_path_cycles(Rules,Graph,SUBCycles,SUBConds),!,
	obtain_conditions_path_cycles_each_subgraph(Rules,Graph,LSUBCycles,LSUBConds),!.

 % un sol cicle
obtain_conditions_path_cycles(Rules,Graph,[Nums],[[RCond|Nums]]):-
	perform_cycle(Graph,[],Nums,Conditions),
	obtain_conds_and_try_rewrite_step(Rules,Conditions,RCond),!.
 % un sol cicle i no es satisf algun nf o no es pot tancar el cicle
obtain_conditions_path_cycles(_,_,[_],[]):-!.    % mai passa - ja s'han eliminat abans
 % varis cicles
obtain_conditions_path_cycles(Rules,Graph,NumCycles,SCycles):-
	selec_nodes_several_cycles(NumCycles,Nodes),!,
	obtain_conditions_path_cycles1(Rules,Graph,Nodes,NumCycles,SCycles),!.
	
obtain_conditions_path_cycles1(_,_,_,[],[]).
obtain_conditions_path_cycles1(Rules,Graph,Nodes,[Nums|NumCycles],[[RCond|Nums]|SCycles]):-
	perform_cycle(Graph,Nodes,Nums,Conditions),
	obtain_conds_and_try_rewrite_step(Rules,Conditions,RCond),!,
	obtain_conditions_path_cycles1(Rules,Graph,Nodes,NumCycles,SCycles),!.	
obtain_conditions_path_cycles1(Rules,Graph,Nodes,[_|NumCycles],SCycles):- % mai passa - ja s'han eliminat
	obtain_conditions_path_cycles1(Rules,Graph,Nodes,NumCycles,SCycles),!.	
		% El cicle es cancel.la perqu no es satisf NF o no es pot tancar el cicle
		% o no pas reescriptura possible

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

perform_cycle(Graph,Nodes,[Num|Nums],Conditions):-
	member21([q(Num,C,Lc),_],Graph),
	refresh(q(Num,C,Lc),T1),
	( 
	  (Nums=[], \+(member21(Num,Nodes)), Nums1=[Num|Nums])
	  ;		% cicles amb 1 sol parell aillat fem 1-1-1
	  (Nums1=Nums)
	),
	construct_cycle(Graph,Nodes,Nums1,[T1],Cycle), 
	obtain_and_renvars_conditions(Cycle,Conditions),!.

% Construeix el cicle respectant els Nodes que no es poden lligar
% perqu formen part de ms d'un cicle.

construct_cycle(Graph,Nodes,Nums,[q(N,C,Lc)],Cycle):-
	refresh(q(N,C,Lc),Qini),	% Qini s el primer abans de fer Cam
	construct_cycle1(Graph,Nodes,Nums,[q(N,C,Lc)],Cam),!,
	((member21(N,Nodes),!, close_cycle(2,Qini,Cam,Cycle))
	 ;
	 (close_cycle(1,q(N,C,Lc),Cam,Cycle))  % q(N,C,Lc) s la instncia que es troba en la 
	),!.	 			 % primera posici del cam desprs de fer Cam. 
					 % Unify  refrescar q(N,C,Lc) pero mant unifs fetes.

%construeix tot el cam sense acabar de tancar el cicle
construct_cycle1(_,_,[],Cycle,Cycle).
construct_cycle1(Graph,Nodes,[Num|Nums],[Q|Cam],Cycle):-
	member212(Num,Nodes),!,
	member21([q(Num,C,Lc),_],Graph),
	unifyQ2(Q,q(Num,C,Lc),Q1),   % unifyQ2 refresca el segon constraint
	refresh(q(Num,C,Lc),Q2),    %el proper connectar amb un nou node NUM Q2	
	construct_cycle1(Graph,Nodes,Nums,[Q2,Q1,Q|Cam],Cycle),!. 
construct_cycle1(Graph,Nodes,[Num|Nums],[Q|Cam],Cycle):-
	member21([q(Num,C,Lc),_],Graph),
	unifyQ2(Q,q(Num,C,Lc),Q1),   % unifyQ2 refresca el segon constraint
	construct_cycle1(Graph,Nodes,Nums,[Q1,Q|Cam],Cycle),!. 


%tanca el cicle [Q|Cam] (est al revs) amb 
%Qini que s el primer constraint (ltim de Cam) refrescat, i
%QiniReal s exactament ltim de Cam (per innermost si K=1 comprovar
% 	que el resultat d'unificar s una instncia del primer. Sin no cicle).
close_cycle(K,Qini,[Q|Cam],[Q2,Q|Cam]):-
	((K=1,!,unifyQ1(Q,Qini,Q2))	
	 ;
	 (unifyQ2(Q,Qini,Q2))
	),!.

%%%%%%%%%%%%%%%%%%%%%%%%%
% LConstraints=[q(N,C,Lc),....]
obtain_and_renvars_conditions(LConstraints,Lc1):-
	obt_lc(LConstraints,[],Lc),
	set_new_vars(Lc),
	renvars_list1(Lc,Lc0),
        del_repes(Lc0,Lc1),!.



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Graph s el subgraf que es correspon a L_SCycles=[SCycles1,...]
% SCycles1 =[ [list-cond|nums],...]
generate_constraints_cycles_each_subgraph(_,_,[],[]):-!.
generate_constraints_cycles_each_subgraph(Rules,Graph,[SCycles|L_SCycles],[SUBSOLCycles|L_SUBSOLCycles]):- 
	generate_constraints_cycles(Rules,Graph,SCycles,SUBSOLCycles),
	generate_constraints_cycles_each_subgraph(Rules,Graph,L_SCycles,L_SUBSOLCycles),!.


% Uni de nums pot ser diferent a tots els nums de Graph perque es poden
%      haver cancel.lat cicles (per NF o no tancar,...), podem tenir subgrafs...
% LORCycles s un OR - diferents maneres de provar el subgraf 

generate_constraints_cycles(_,_,[],[]):-!.
generate_constraints_cycles(Rules,Graph,SCycles,[NumsGraph,StrictCycles,NonStrictCycles,LORCycles]):-
	findall(N,member22([q(N,_>_,_),_],Graph),LStricts),
	mark_strict_cycles_1_or_more(LStricts,SCycles,SCyclesMark),!,  % [ [ge,list-conds|nums],[g,list-conds|nums]...]
	((
	   member21([ge|_],SCyclesMark),!,
	   generate_constraints_b(Rules,Graph,SCyclesMark,LORCycles)
	 )
	 ;
	 (
	   generate_constraints_a(Rules,Graph,SCyclesMark,LORCycles)  %tots els cicles sn estrictes
	 )
	),
	separ_strict_nonstrict_cycles(SCycles,Graph,StrictCycles,NonStrictCycles),   
	obtain_numsGraph_SCycles(SCycles,NumsGraph),!.
generate_constraints_cycles(_,_,_,[]). % no hi ha cap estricte

% Degut a no NF, no tancar cicle,... alguns cicles es poden haver cancel.lat
% i podria no quedar cap estricte.
mark_strict_cycles_1_or_more(LStricts,SCycles,SCyclesMark):-
	mark_strict_cycles(LStricts,SCycles,SCyclesMark),!,
	member21([g,_|_],SCyclesMark).

mark_strict_cycles(_,[],[]).
mark_strict_cycles(LStricts,[[Cond|Nums]|SCycles],[[Kind,Cond|Nums]|SCyclesMark]):-
	( ( inter(LStricts,Nums,[_|_]),!,Kind=g )
	  ;
	  ( Kind=ge )
	),!,
	mark_strict_cycles(LStricts,SCycles,SCyclesMark),!.


%%%%%%%%%%%%%%%%%%%%%%%%%%%

obtain_numsGraph_SCycles(SCycles,NumsGraph):-
	map(tail,SCycles,Nums),
	l_union(Nums,NumsGraph),!. 


%%%%%%%%%%%%%%%%%%%%%%%%%%%

generate_constraints_a(Rules,Graph,SCyclesMark,LCycles):-
	obtain_option_3(Rules,Graph,SCyclesMark,Opt3),
	Opt3=[_|_],!,
	modify_SCycles_removing_condsOpt3(SCyclesMark,Opt3,SCyclesMark1),
	obtain_all_nums(SCyclesMark1,[],LNums),
	findall([Conds|Nums],member22([g,Conds|Nums],SCyclesMark1),LCycles0),
	LCycles=[['a1',LNums,LCycles0]|Opt3],!.
generate_constraints_a(_,_,SCyclesMark,[['a1',LNums,LCycles]]):-
	obtain_all_nums(SCyclesMark,[],LNums),
	findall([Conds|Nums],member22([g,Conds|Nums],SCyclesMark),LCycles),!.
	

modify_SCycles_removing_condsOpt3(SCyclesMark,Opt3,SCyclesMarkOut):-
	findall(Lconds,member22([_,_,Lconds|_],Opt3),LLconds),
	l_union(LLconds,LcondsOpt3),
	remove_LcondsOpt3(SCyclesMark,LcondsOpt3,SCyclesMarkOut),!.

remove_LcondsOpt3([],_,[]):-!.
remove_LcondsOpt3([[g,Lconds|R]|SCyclesMark],LcondsOpt3,[[g,Lconds1|R]|SCyclesMarkOut]):-
	difer(Lconds,LcondsOpt3,Lconds1),
	remove_LcondsOpt3(SCyclesMark,LcondsOpt3,SCyclesMarkOut),!.




obtain_all_nums([],L,L).
obtain_all_nums([[_,_|R]|SCyclesMark],LNums,LNumsOut):-
	union(R,LNums,LNums1),
	obtain_all_nums(SCyclesMark,LNums1,LNumsOut),!.
	
%%%%%%%%%%%%%%%%%%%%%%%%%%%

generate_constraints_b(Rules,Graph,SCyclesMark,LCycles):-
	obtain_option_1(SCyclesMark,Opt1),
	obtain_option_2('a2',SCyclesMark,Opt2),
	obtain_option_3(Rules,Graph,SCyclesMark,Opt3),
	( (member21([b,_,_,[]],Opt3),!,Opts=[Opt2,Opt3] )  % per no repetir solucions
	  ;
	  (Opts=[Opt1,Opt2,Opt3])
	),
	l_concat(Opts,LCycles),!.
	
% opci que cancel.la tots els cicles estrictes
obtain_option_1(SCyclesMark,[]):-
	member21([g,[]|_],SCyclesMark),!.
obtain_option_1(SCyclesMark,[['a1',[],LCondsOut]]):-  %LCondsOut=[[l-conds-1],[l-conds-2],...]]
	findall(Conds,member22([g,Conds|_],SCyclesMark),LConds),
	simplify_LConds(LConds,LConds1),
	map(list1,LConds1,LCondsOut),!.  % per fer ms homogeni amb solucions que tenen nums
	
simplify_LConds([],[]).
simplify_LConds([L|LConds],LCondsOut):-
	simplify_LConds1(L,LConds,LConds1,RL),
	simplify_LConds(LConds1,LCondsOut1),
	concat312(RL,LCondsOut1,LCondsOut),!.

simplify_LConds1(L,[],[],[L]):- !.
simplify_LConds1(L,[L1|Rest],[L1|Rest],[]):-
	difer(L1,L,[]),!.
simplify_LConds1(L,[L1|Rest],R,Rx):-
	difer(L,L1,[]),!,
	simplify_LConds1(L,Rest,R,Rx),!.
simplify_LConds1(L,[L1|Rest],[L1|Rest1],Rx):-
	simplify_LConds1(L,Rest,Rest1,Rx),!.


% opci que no considera Conds per cap cicle: Tots constraints i llista cicles estrictes
% Si tots els cicles sn 'g' es crida amb K=a1
% Si algun 'ge' es crida amb K=a2
obtain_option_2(K,SCyclesMark,[[K,LNums,LNumsCycleStricts]]):-
	obtain_all_nums(SCyclesMark,[],LNums),
	findall([[]|Nums],member22([g,_|Nums],SCyclesMark),LNumsCycleStricts),!.
	
% cancel.la un cicle A i obt nous subgrafs que cal provar de manera independent
% afegint a tots les condicions de cancel.lar el cicle A
% Si no hi ha cap cicle que cancel.lant-lo obtinguem subgrafs (o un i altres tots ge)
% aleshores no hi ha opci 3. 
obtain_option_3(_,_,_,[]):-
	\+(deeper_analysis),!.
obtain_option_3(Rules,Graph,SCyclesMark,Option_3):-
	obtenir_connexions_cycles(SCyclesMark,IdCycles,ConnCycles),
	obtenir_llista_condicions(ConnCycles,Llista_conds),
	Llista_conds=[_|_],
	grups_cicles_mateixes_condicions(Llista_conds,ConnCycles,Llista_Lconds_cicles),
  %Llista_Lconds_cicles=[[Lconds1,Lnumcicles],[Lconds2,Lnumcicles],...]
	setof(Opt,cancel_cycle_new_subgraphs(Rules,Graph,IdCycles,Llista_Lconds_cicles,ConnCycles,Opt),LOptions),	
		% si LOptions=[] falla  , 
		% LOptions=[[Lcond1, Cycs1, [[K,Lnums11,Lstr11,Lnostr11],...]],
		%           [Lcond2, Cycs2, [[K,Lnums21,Lstr21,Lnostr21,...]],...]	
		% Si obtenim soluci trivial tenim LOptions=[...,[Lcond, Cycs, []],...]
	selec_best_options(LOptions,Option), % Option s un Or : possibles maneres de cancel.lar cicles
	enumerate_cons(Option,OptionEn),     %			 amb una condici
	map(cons,b,OptionEn,Option_3),!.
obtain_option_3(_,_,_,[]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%SCyclesMark=[[Kind,list-conds|nums],...]
obtenir_connexions_cycles(SCyclesMark,LIdCycles,ConnCycles):-
	enumerate(SCyclesMark,LIdCycles),  %LIdCycles=[[1,[Kind,list-conds|nums]],[2,[...]],...]
	connexions_cycles(LIdCycles,LIdCycles,ConnCycles),!.

%ConnCycles=[[q(idCycle,_Kind_,Lc),ListConnexOtherCycles],...]
connexions_cycles([],_,[]).
connexions_cycles([[N,[K,Lc|Nums]]|LIdCycles],AllIdCycles,[[q(N,C,Lc),[N|LM]]|ConnCycles]):-
	((K=g,!,C=..[>|[*,*]])
	 ;
	 (C=..[>=|[*,*]])
	),
	findall(M,connection(N,Nums,AllIdCycles,M),LM),!,
	connexions_cycles(LIdCycles,AllIdCycles,ConnCycles),!.
		
connection(N,Nums,[[M,[_,_|L]]|_],M):-
	\+(N=M),
	inter(Nums,L,[_|_]).
connection(N,Nums,[_|AllIdCycles],M):-
	connection(N,Nums,AllIdCycles,M).


obtenir_llista_condicions(ConnCycles,Llista_conds):-
	findall(Lc,member22([q(_,_,Lc),_],ConnCycles),Llista1),
	l_union(Llista1,Llista_conds),!.

grups_cicles_mateixes_condicions(Llista_conds,ConnCycles,Llista_Lconds_cicles):-
	map(grups_cicles_condicio,ConnCycles,Llista_conds,Llista_cond_cicles),
	agrupar_condicions_mateixos_cicles(Llista_cond_cicles,[],Llista_Lconds_cicles),!.

grups_cicles_condicio(Graph,Cond,[Cond,Lcicles]):-
	findall(N,shared_cond_cycles(Cond,Graph,N),Lcicles),!.
	
shared_cond_cycles(Cond,Graph,N):-
	member22([q(N,_,Lc),_],Graph),
	member21(Cond,Lc).

agrupar_condicions_mateixos_cicles([],L,L).
agrupar_condicions_mateixos_cicles([[Cond,Lcicles]|Llista_cond_cicles],L,Llista_Lconds_cicles):-
	delete32([Lconds,Lcicles],L,L1),!,
	agrupar_condicions_mateixos_cicles(Llista_cond_cicles,[[[Cond|Lconds],Lcicles]|L1],Llista_Lconds_cicles),!.
agrupar_condicions_mateixos_cicles([[Cond,Lcicles]|Llista_cond_cicles],L,Llista_Lconds_cicles):-
	agrupar_condicions_mateixos_cicles(Llista_cond_cicles,[[[Cond],Lcicles]|L],Llista_Lconds_cicles),!.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%  IdCycles=[[idcycle,[Kind,list-conds|nums]],[2,[...]],...] 
%  ConnCycles=GraphCycles=[[q(idCycle,_Kind_,Lc),ListConnexOtherCycles],...]
%  LlistaCondsCycles=[[Lconds1,Lcicles1],[Lconds2,Lcicles2],...]
%
%  LcancelCycs=[LcancelStricts,LcancelNoStricts]
cancel_cycle_new_subgraphs(Rules,IniGraph,IdCycles,[[Lconds,Lcycles]|_],GraphCycles,
						[Lconds,LcancelCycs,LNewSubgrafsConds]):-
	remove_list_N_GraphCycles(Lcycles,GraphCycles,Graph1),
	simplify_graph(Graph1,LSubgraphsLconds),
	check_interesting_solution(Lcycles,Graph1,LSubgraphsLconds),
 	obtain_cancelled_cycles(Lcycles,IdCycles,LcancelCycs),
	treat_1_CondSubgraphs(Rules,IniGraph,IdCycles,LSubgraphsLconds,LNewSubgrafsConds).
cancel_cycle_new_subgraphs(Rules,IniGraph,IdCycles,[_|LlistaConds],GraphCycles,Opt):-
	cancel_cycle_new_subgraphs(Rules,IniGraph,IdCycles,LlistaConds,GraphCycles,Opt).

%%%%%%%
remove_list_N_GraphCycles([],Graph,Graph).
remove_list_N_GraphCycles([C|LcycleShareCond],Graph,GraphOut):-
	remove_N_GraphCycles(C,Graph,Graph1),!,
	remove_list_N_GraphCycles(LcycleShareCond,Graph1,GraphOut),!.

remove_N_GraphCycles(_,[],[]).
remove_N_GraphCycles(N,[[q(N,_,_),_]|Rest],Rest1):-!,
	remove_N_GraphCycles(N,Rest,Rest1),!.
remove_N_GraphCycles(N,[[X,LM]|Rest],[[X,LM1]|Rest1]):-
	delete0312(N,LM,LM1),!,
	remove_N_GraphCycles(N,Rest,Rest1),!.

%%%%%%%
%  IdCycles=[[idcycle,[Kind,list-conds|nums]],[2,[...]],...] 
%  StrictCycs-NonStrictCycs=[LNums1,LNums2,...] 
obtain_cancelled_cycles(Lcycles,IdCycles,[StrictCycs,NonStrictCycs]):-
	obtain_cycles_Lcycles(Lcycles,IdCycles,[],StrictCycs,[],NonStrictCycs),!.

% No cal retornar les condicions dels cicles cancel.lats
obtain_cycles_Lcycles([],_,StrictCycs,StrictCycs,NonStrictCycs,NonStrictCycs):-!.
obtain_cycles_Lcycles([IdCyc|Lcycles],IdCycles,S,StrictCycs,NS,NonStrictCycs):-
	member21([IdCyc,[K,_|Nums]],IdCycles),!,
	( ( K=g,!, obtain_cycles_Lcycles(Lcycles,IdCycles,[Nums|S],StrictCycs,NS,NonStrictCycs) )
	  ;
	  (obtain_cycles_Lcycles(Lcycles,IdCycles,S,StrictCycs,[Nums|NS],NonStrictCycs) )
	),!.

%%%%%%%
check_interesting_solution([_],Graph,[G]):-!,  % comprovem que tota condici cancel.li ms d'un cicle
	length21(Graph,N1),		       % o divideixi en subgrafs 
	length21(G,N2),
	N2<N1,!.
check_interesting_solution(_,_,_).


%%%%%%%
treat_1_CondSubgraphs(_,_,_,[],[]):-!.
treat_1_CondSubgraphs(Rules,IniGraph,IdCycles,[Sub|L_Subgraphs],SOut):-
	treat_1_Subgraph(Rules,IniGraph,IdCycles,Sub,LS1),!,
	treat_1_CondSubgraphs(Rules,IniGraph,IdCycles,L_Subgraphs,Rest),
	concat312(LS1,Rest,SOut),!.
treat_1_CondSubgraphs(Rules,IniGraph,IdCycles,[_|L_Subgraphs],Rest):-  % subgraf s'ha cancel.lat
	treat_1_CondSubgraphs(Rules,IniGraph,IdCycles,L_Subgraphs,Rest),!.

% Sub=[[q(Ncycle,*K*,Lc),Conn_cycles],...]
% falla quan desprs de construir Lc de Sub ens quedem amb cap cicle estricte: Sub es cancel.lar
treat_1_Subgraph(Rules,IniGraph,IdCycles,Sub,L_SolSub):-
	findall(N,member22([q(N,_,_),_],Sub),LNCyclesSubgraph),
	obtain_simple_cycles_Sub(IdCycles,LNCyclesSubgraph,SimpleCyclesSubgraph),
	obtain_subgraphs_and_conditions_path_cycles(Rules,IniGraph,SimpleCyclesSubgraph,LSUBCycles),
		% LSUBCycles=[SCycles1,SCycles2,...].
		% SCyclesI= list of [list-conditions,nums-cycles]. e.g. [[cond1,cond2,...],1,3,4] 
		% nums-cycles contain all stricts and non-stricts
	LSUBCycles=[_|_],	
	obtain_constraint_solve_each_subgraph(IniGraph,LSUBCycles,L_SolSub),!. 

obtain_simple_cycles_Sub(_,[],[]).
obtain_simple_cycles_Sub(IdCycles,[Ncycle|LNcyclesSubgraph],[Lnums|SimpleCyclesSubgraph]):-
	member21([Ncycle,[_,_|Lnums]],IdCycles),
	obtain_simple_cycles_Sub(IdCycles,LNcyclesSubgraph,SimpleCyclesSubgraph),!.


obtain_constraint_solve_each_subgraph(_,[],[]):-!.
obtain_constraint_solve_each_subgraph(IniGraph,[SUBCycles|LSUBCycles],[SOLSUBCycles|L_SolSub]):-
	obtain_constraint_solve_subgraph(IniGraph,SUBCycles,SOLSUBCycles),!,
	obtain_constraint_solve_each_subgraph(IniGraph,LSUBCycles,L_SolSub),!.
obtain_constraint_solve_each_subgraph(IniGraph,[_|LSUBCycles],L_SolSub):-
	obtain_constraint_solve_each_subgraph(IniGraph,LSUBCycles,L_SolSub),!.
				 % subgraf trivial, cap estricte
	

% falla si no hi ha cap estricte
% Si algun ge retorna llista-constraints, llista-cicles-estrictes (amb cond=[])
% Si cap ge retorna llista-constraints, llista-cicles-estrictes (amb cond)
obtain_constraint_solve_subgraph(IniGraph,SCycles,[K,LNums,LCycles,LCyclesNoStrict]):-
	findall(N,member22([q(N,_>_,_),_],IniGraph),LStricts),
	mark_strict_cycles_1_or_more(LStricts,SCycles,SCyclesMark),!,  
			% [ [ge,list-conds|nums],[g,list-conds|nums]...]
	obtain_all_nums(SCyclesMark,[],LNums),
	((
	   member21([ge|_],SCyclesMark),!, K='a2',			
	   findall([[]|Nums],member22([g,_|Nums],SCyclesMark),LCycles),
	   findall([[]|Nums],member22([ge,_|Nums],SCyclesMark),LCyclesNoStrict)
	 )
	 ;
	 (
	   K='a1',
   	   findall([Conds|Nums],member22([g,Conds|Nums],SCyclesMark),LCycles),
	   LCyclesNoStrict=[]
	 )
	),!.
	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% LOptions=[[Lcond1, Cycs1, [[K,Lnums11,Lstr11,Lnostr11],...]],
%           [Lcond2, Cycs2, [[K,Lnums21,Lstr21,Lnostr21,...]],...]	
% Si obtenim soluci trivial tenim LOptions=[...,[Lcond, Cycs, []],...]
% Option com LOptions per noms amb els de OptVal
	
selec_best_options(LOptions,Option):-   % Option=[[Lcond1,Cycs1,[]],[Lcond2,Cycs2,[]],...]
	setof([Lcond,Cycs,[]],member22([Lcond,Cycs,[]],LOptions),Option),!. 
selec_best_options(LOptions,Option):-   % cap Or cont una soluci trivial
	map(evaluate_each_option,LOptions,LOptionsVal),
	sort(LOptionsVal,[[OptVal|_]|_]),	% OptVal t minim NumConstraints
	selec_options_or_Val(OptVal,LOptionsVal,Option),!.  
		% del OR noms agafa els que tenen NumConstraints de la millor opci.

get_sols_trivials([Or_Opt|_],[Lcond,[]]):-
	member22([Lcond,[]],Or_Opt).
get_sols_trivials([_|LOptions],Opt):-
	get_sols_trivials(LOptions,Opt).


evaluate_each_option([Lcond,Cycs,Subgrafs],[Val,Lcond,Cycs,Subgrafs]):-
	count_number_constraints(Subgrafs,0,Val),!.
		
count_number_constraints([],N,N).
count_number_constraints([[_,Lnums,_,_]|Subgrafs],N,NumConstraints):-
	long(Lnums,N,N1),
	count_number_constraints(Subgrafs,N1,NumConstraints),!.


selec_options_or_Val(NumConstraints,LOptionsVal,Option):-
	findall(X,member22([NumConstraints|X],LOptionsVal),Option),!.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


% unifica la part dreta del primer parell amb l'esquerra del segon 
% (prviament refresca el segon constraint)
unifyQ1(q(_,Q1,_),Cons2,q(N,Q22,Lc2)):-  Q1=..[_|[_,T1]], 
				refresh(Cons2,q(N,Q22,Lc2)), 
				Q22=..[_|[S2,_]],!,
				check_unify(T1,S2,Cons2).

% si s innermost, cal unificar amb occur check i a ms una vegada unificat amb la
% versi renombrada recomprovar que segueix essent una instncia del primer terme
%(innermost-->pot ser que al unificar s'hagi lligat alguna variable del primer terme)

check_unify(S,T,q(_,Q,_)):-innermost,!,unifica(S,T), 
			  Q=..[_|[Sa,_]], refreshterm(Sa,S1),correct(S1,S),!.
check_unify(S,T,_):-S=T,!.
			
% correct: comprova que S i T unifiquen per sense retornar unificats.
correct(S,T):-prova(S,T),!,fail.
correct(_,_).
prova(S,T):-unifica(S,T),!,fail.
prova(_,_).


unifyQ2(q(_,Q1,_),Cons2,q(N,Q22,Lc2)):-  Q1=..[_|[_,T1]], 
				refresh(Cons2,q(N,Q22,Lc2)), 
				Q22=..[_|[S2,_]],!,
				unify0(T1,S2).


unifica(S,T):- nounif(S,T),!,fail.
unifica(S,T):- unif(S,T),!.

nounif(S,T):-S=T,!,fail.
nounif(_,_).

%%----------------------------
%% unificaci amb occur check
%%----------------------------

unif(T1,T2):-var(T1), var(T2),!, T1=T2,!.
unif(T1,T2):-var(T1), occurs_chk1(T2,T1),!,fail.
unif(T1,T2):-var(T1),!, T1=T2,!.
unif(T1,T2):-var(T2), occurs_chk1(T1,T2),!,fail.
unif(T1,T2):-var(T2),!, T1=T2,!.
unif(T1,T2):-T1=..[F|Args1], T2=..[F|Args2],
		uniflist(Args1,Args2),!.
uniflist([],[]):-!.
uniflist([X|R1],[Y|R2]):-unif(X,Y),!, uniflist(R1,R2).



%--------------------------------------------------
% funcions per refrescar variables
%--------------------------------------------------

% refrescar totes les variables d'un constraint
refresh(q(N,C,Lc),q(N,C1,Lc1)):- C=..[Op|[S,T]],
			llvars(S,[],LV1), llvars(T,LV1,LV),llvarsnews(LV,LVN),!,  
			replace(LVN,S,S1),replace(LVN,T,T1),
			C1=..[Op|[S1,T1]], 
			replace(LVN,Lc,Lc1).

% refrescar un terme
refreshterm(S,S1):- llvars(S,[],LV), llvarsnews(LV,LVN),!, replace(LVN,S,S1),!.

% refrescar tot un cam.
refreshcam(Cam,Camnou):-llvars(Cam,[],LV), llvarsnews(LV,LVN), 
			replace(LVN,Cam,Camnou),!.

% afegeix totes les variables d'un terme o llista TL a una llista donada (2n. arg).

llvars(TL,LI,LV):-llvars1(TL,LI,LO),reverse1(LO,LV).

llvars1(T,L,L):-ground(T),!.
llvars1(T,L,L):-member212(T,L),!.
llvars1(T,L,[T|L]):-var(T),!.
llvars1([],L,L):-!.
llvars1([X|R],L1,L):-!, llvars1(X,L1,L2), llvars1(R,L2,L),!.
llvars1(T,L1,L):-T=..[_|Args],llvars1(Args,L1,L),!.



% donada una llista de variables LV, retorna una llista de llistes <var,varnova>
llvarsnews([],[]):-!.
llvarsnews([X|R],[[X,Y]|R1]):-llvarsnews(R,R1),!.


% Donada una llista (1r.arg) de parells <var,varnova>, renombrar en el terme T
replace([],T,T):-!.
replace([[X,Y]|R],T,T1):-var(T),!,replace_all(X,Y,T,T0),replace(R,T0,T1),!.
replace(_,[],[]):-!.
replace(LV,[Term|RTerm],[T1|R1]):-replace(LV,Term,T1), replace(LV,RTerm,R1),!.
replace([[X,Y]|R],Term,T1):- replace_all(X,Y,Term,T), replace(R,T,T1),!.

% substitueix totes les ocurrncies de X per Y a T(qualsevol estruct. o llista).
replace_all(X,Y,T,Y):-X==T,!.
replace_all(_,_,T,T):-var(T),!.
replace_all(_,_,[],[]):-!.
replace_all(X,Y,[Z|R],[Z1|R1]):-replace_all(X,Y,Z,Z1),
			   	replace_all(X,Y,R,R1),!.
replace_all(X,Y,T,T1):-	T=..[F|Args],
			replace_all(X,Y,Args,Args1),
			T1=..[F|Args1],!.

